package io.scanbot.sdk.ui.camera

import android.annotation.SuppressLint
import android.app.Activity
import android.content.Context
import android.content.pm.ActivityInfo
import android.graphics.*
import android.hardware.SensorManager
import android.hardware.camera2.CameraCharacteristics
import android.hardware.camera2.CameraManager
import android.hardware.camera2.CameraMetadata
import android.hardware.camera2.CaptureRequest
import android.hardware.display.DisplayManager
import android.media.ImageReader
import android.util.AttributeSet
import android.util.Size
import android.view.*
import android.widget.FrameLayout
import androidx.camera.camera2.interop.Camera2Interop
import androidx.camera.core.*
import androidx.camera.core.Camera
import androidx.camera.core.impl.ImageOutputConfig
import androidx.camera.lifecycle.ProcessCameraProvider
import androidx.camera.view.PreviewView
import androidx.core.content.ContextCompat
import androidx.core.view.children
import androidx.core.view.isVisible
import androidx.lifecycle.Lifecycle
import androidx.lifecycle.LifecycleObserver
import androidx.lifecycle.LifecycleOwner
import androidx.lifecycle.OnLifecycleEvent
import io.scanbot.sap.SdkFeature
import io.scanbot.sdk.base.R
import io.scanbot.sdk.camera.*
import io.scanbot.sdk.exceptions.camera.CameraConfigurationException
import io.scanbot.sdk.security.SapSingleton
import io.scanbot.sdk.ui.BasePolygonView
import io.scanbot.sdk.ui.camera.util.*
import io.scanbot.sdk.ui.camera.util.PreviewImageUtil.convertYUV420toNV21
import io.scanbot.sdk.ui.camera.util.ScaleTypeTransform.rotationDegreesFromSurfaceRotation
import io.scanbot.sdk.util.log.LoggerProvider
import java.util.concurrent.Executors
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.atomic.AtomicInteger
import kotlin.math.abs
import kotlin.math.max
import kotlin.math.min

/**
 * Wrapper view for CameraX
 *
 * @BETA CameraX implementation of the CameraView is based on the beta-version of CameraX library so it might be unstable.
 */
class ScanbotCameraXCView(context: Context, attrs: AttributeSet?) : FrameLayout(context, attrs),
    IScanbotCameraView, UiZoomDelegate,
    FinderViewAttributeHandler by FinderViewAttributeHandlerImpl(context, attrs) {

    constructor(context: Context) : this(context, null)

    private var forceMaxSnappingSize: Boolean = false

    var supportedPreviewSizes: Array<Size> = emptyArray()
    var supportedPictureSizes: Array<Size> = emptyArray()

    private var externalCaptureCallback: CaptureCallback? = null

    var lifecycleOwner: LifecycleOwner? = null
        get() {
            return if (field == null) {
                context as? LifecycleOwner
            } else field
        }

    private var customPreviewFrameSize: Size? = null
    private var customAnalyzerFrameSize: Size? = null
    private var customPictureSize: Size? = null

    private val logger = LoggerProvider.logger

    private val previewView: PreviewView

    private lateinit var cameraProvider: ProcessCameraProvider
    private var cameraManager: CameraManager? = null

    private var autoFocusOnTouch: Boolean = true
    private var touchFocusPaint: Paint
    private var touchRect: Rect? = null

    private var isPictureCapturedAutomatically: Boolean = false

    private var resolution: Size? = null
    private var meteringPointFactory: PreviewViewMeteringPointFactory? = null

    private val pinchToZoomEnabled: AtomicBoolean = AtomicBoolean(true)
    private var scaleGestureDetector: ScaleGestureDetector? = null
    private var zoomRange: ZoomRange = ZoomRange(0f, 1f)
    private var initialZoomLevel: Float = -1f
    private var physicalZoom: Float = -1f

    private var displayId: Int = -1
    private var cameraXLensFacing: Int = CameraSelector.LENS_FACING_BACK
    private var scanbotCameraModule: CameraModule = CameraModule.BACK
    private val lockMinFocusDistance: AtomicBoolean = AtomicBoolean(false)

    private var imageCaptureUseCase: ImageCapture? = null
    private var previewUseCase: Preview? = null

    private var imageAnalyzer: ImageAnalysis? = null

    private val compoundImageAnalyzer: CompoundAnalyzer = CompoundAnalyzer()
    private var camera: Camera? = null

    private val stateCallbacks: MutableSet<CameraStateCallbackWrapper> = LinkedHashSet()
    private val pictureCallbacks: MutableSet<BasePictureCallback> = LinkedHashSet()
    private val takePictureCallbacks: MutableSet<CameraTakePictureCallback> = LinkedHashSet()
    private var cameraOpenCallback = CameraOpenCallback.NULL

    private var previewMode: CameraPreviewMode = CameraPreviewMode.FILL_IN
    private var isOrientationLocked = false
    private var isOrientationHardLocked = false

    private var finderInnerThresholdPx = 0
    private var finderOuterThresholdPx = 0
    private var currentFinderRectF: List<PointF> = emptyList()

    override var delayAfterFocusCompleteMs: Long = 20

    private val displayManager by lazy {
        context.getSystemService(Context.DISPLAY_SERVICE) as DisplayManager
    }
    private var snapAnimationView: SnapFlashView

    /** Blocking camera operations are performed using this executor */
    private val cameraExecutor = Executors.newSingleThreadExecutor()
    private val analyzerExecutor = Executors.newSingleThreadExecutor()
    private val frameAnalyzerExecutor = Executors.newSingleThreadExecutor()

    /**
     * We need a display listener for orientation changes that do not trigger a configuration
     * change, for example if we choose to override config change in manifest or for 180-degree
     * orientation changes.
     */
    private val displayListener = object : DisplayManager.DisplayListener {
        override fun onDisplayAdded(displayId: Int) = Unit
        override fun onDisplayRemoved(displayId: Int) = Unit

        @SuppressLint("UnsafeOptInUsageError")
        override fun onDisplayChanged(displayId: Int) = this@ScanbotCameraXCView.let { view ->
            if (displayId == view.display.displayId) {
                val newRotation = view.display.rotation
                logger.d(TAG, "Rotation changed: $newRotation")
                imageCaptureUseCase?.targetRotation = newRotation
                previewUseCase?.targetRotation = newRotation
                imageAnalyzer?.targetRotation = newRotation
            }
        }
    }

    private var onOrientationChangeListener: OnOrientationChange? = null

    private var isPreviewResumed = false

    init {
        inflate(context, R.layout.scanbot_camerax_view, this)

        touchFocusPaint = Paint()
        touchFocusPaint.color = ContextCompat.getColor(context, android.R.color.white)
        touchFocusPaint.style = Paint.Style.STROKE
        touchFocusPaint.strokeWidth = resources.getDimension(R.dimen.touch_focus_polygon_width)
        touchFocusPaint.isAntiAlias = true

        finderInnerThresholdPx = context.resources.getDimensionPixelSize(R.dimen.default_finder_inner_threshold)
        finderOuterThresholdPx = context.resources.getDimensionPixelSize(R.dimen.default_finder_outer_threshold)

        if (!isInEditMode) {
            onOrientationChangeListener = OnOrientationChange(context)
        }

        previewView = findViewById(R.id.camera_preview_view)

        // Workaround for the preview issue on Samsung S6 https://github.com/android/camera-samples/issues/212
        previewView.implementationMode = PreviewView.ImplementationMode.PERFORMANCE
        if (context is LifecycleOwner) {
            previewView.previewStreamState.observe(context as LifecycleOwner) { state ->
                when (state) {
                    PreviewView.StreamState.STREAMING -> {
                        if (scanbotCameraModule == CameraModule.FRONT) {
                            previewView.scaleX = -1F
                        }
                        if (isPreviewResumed.not()) {
                            isPreviewResumed = true
                            cameraOpenCallback.onCameraOpened()
                        }
                    }
                    PreviewView.StreamState.IDLE, null -> {
                        isPreviewResumed = false
                    }
                }
            }
        }

        if (!isInEditMode) {
            cameraManager = context.getSystemService(Context.CAMERA_SERVICE) as CameraManager
            initSupportedPreviewPictureSizes()
        }

        snapAnimationView = findViewById(R.id.flashView)
        snapAnimationView.isVisible = false
    }

    override fun onAttachedToWindow() {
        super.onAttachedToWindow()

        onOrientationChangeListener?.resetOrientation()
        // Every time the orientation of device changes, update rotation for use cases
        displayManager.registerDisplayListener(displayListener, null)

        // Wait for the views to be properly laid out
        post {
            if (previewView.display != null) {
                // Keep track of the display in which this view is attached
                displayId = previewView.display.displayId

                this.previewView.scaleType = when (previewMode) {
                    CameraPreviewMode.FIT_IN -> PreviewView.ScaleType.FIT_CENTER
                    CameraPreviewMode.FILL_IN -> PreviewView.ScaleType.FILL_CENTER
                }

                // Bind use cases
                setUpCamera()
                setUpPinchToZoom()
            }
        }
    }

    override fun onDetachedFromWindow() {
        super.onDetachedFromWindow()
        snapAnimationView.cancel()
        // Shut down our background executor
        cameraExecutor.shutdown()
        analyzerExecutor.shutdown()
        frameAnalyzerExecutor.shutdown()

        // Unregister the broadcast receivers and listeners
        displayManager.unregisterDisplayListener(displayListener)
        onOrientationChangeListener?.disable()

        previewUseCase = null
        imageCaptureUseCase = null
        imageAnalyzer = null
    }

    override fun <T : FrameHandler> getAttachedFrameHandler(clazz: Class<T>): T? {
        logger.logMethod()
        return compoundImageAnalyzer.getAttachedFrameHandler(clazz)
    }

    /**
     * Used to add frame handler that will be called on each next frame
     */
    @Synchronized
    override fun addFrameHandler(frameHandler: FrameHandler) {
        compoundImageAnalyzer.addFrameHandler(frameHandler)
    }

    @Synchronized
    override fun removeFrameHandler(frameHandler: FrameHandler) {
        compoundImageAnalyzer.removeFrameHandler(frameHandler)
    }

    override fun setPreviewMode(previewMode: CameraPreviewMode) {
        this.previewMode = previewMode
    }

    override fun takePicture(acquireFocus: Boolean) {
        takePicture(acquireFocus, false)
    }

    override fun takePicture(acquireFocus: Boolean, isCapturedAutomatically: Boolean) {
        logger.logMethod()
        if (lifecycleOwner?.lifecycle?.currentState?.isAtLeast(Lifecycle.State.STARTED) == false) {
            return
        }
        isPictureCapturedAutomatically = isCapturedAutomatically
        if (!SapSingleton.getInstance().checkLicenseStatus(SdkFeature.NoSdkFeature)) {
            notifyTakePictureCancelled()
            return
        }

        if (acquireFocus) {
            tryAutoFocus(object : AutofocusCallback {
                override fun onAutoFocusCompleted() {
                    logger.logMethod()
                    postDelayed({
                        tryTakePicture()
                    }, delayAfterFocusCompleteMs)
                }
            })
        } else {
            tryTakePicture()
        }
    }

    private fun tryTakePicture() {
        // Get a stable reference of the modifiable image capture use case
        imageCaptureUseCase?.let { imageCapture ->
            if (cameraExecutor.isShutdown) {
                return
            }
            // Setup image capture listener which is triggered after photo has been taken
            imageCapture.takePicture(cameraExecutor, object : ImageCapture.OnImageCapturedCallback() {
                override fun onCaptureSuccess(image: ImageProxy) {
                    if (cameraExecutor.isShutdown) {
                        return
                    }

                    post {
                        snapAnimationView.playSelf()
                        externalCaptureCallback?.onImageCaptured()
                    }
//
//                    cameraExecutor.execute(
//                        JPEGImageSaver(
//                            image,
//                            cameraExecutor,
//                            object : JPEGImageSaver.OnImageProcessedCallback {
//                                override fun onImageSaved(outputImage: ByteArray, rotationDegrees: Int) {
//                                    pictureCallbacks.forEach {
//                                        it.onPictureTakenInternal(
//                                            outputImage,
//                                            rotationDegrees,
//                                            currentFinderRectF,
//                                            isPictureCapturedAutomatically
//                                        )
//                                    }
//                                }
//
//                                override fun onError(saveError: JPEGImageSaver.SaveError?, message: String?, cause: Throwable?) {
//                                    logger.e(TAG, "Photo capture failed: $message")
//                                    logger.logException(cause)
//                                    notifyTakePictureCancelled()
//                                }
//                            })
//                    )
                }

                override fun onError(exception: ImageCaptureException) {
                    logger.e(TAG, "Photo capture failed: ${exception.message}")
                    logger.logException(exception)
                    notifyTakePictureCancelled()
                }
            })
        }
    }

    override fun addPictureCallback(pictureCallback: BasePictureCallback) {
        logger.logMethod()
        pictureCallbacks.add(pictureCallback)
    }

    override fun removePictureCallback(pictureCallback: BasePictureCallback) {
        logger.logMethod()
        pictureCallbacks.remove(pictureCallback)
    }

    override fun setCameraOpenCallback(cameraOpenCallback: CameraOpenCallback) {
        this.cameraOpenCallback = cameraOpenCallback
    }

    override fun addCameraStateCallback(cameraStateCallback: CameraStateCallback) {
        logger.logMethod()
        val cameraStateCallbackWrapper = CameraStateCallbackWrapper(cameraStateCallback)
        lifecycleOwner?.lifecycle?.addObserver(cameraStateCallbackWrapper)
        stateCallbacks.add(cameraStateCallbackWrapper)
    }

    override fun setCaptureCallback(captureCallback: CaptureCallback?) {
        externalCaptureCallback = captureCallback
    }

    override fun removeCameraStateCallback(cameraStateCallback: CameraStateCallback) {
        logger.logMethod()
        stateCallbacks.filter { it.stateCallback === cameraStateCallback }.forEach { callbackWrapper ->
            lifecycleOwner?.lifecycle?.removeObserver(callbackWrapper)
            stateCallbacks.remove(callbackWrapper)
        }
    }

    override fun addTakePictureCallback(cameraTakePictureCallback: CameraTakePictureCallback) {
        logger.logMethod()
        takePictureCallbacks.add(cameraTakePictureCallback)
    }

    override fun removeTakePictureCallback(cameraTakePictureCallback: CameraTakePictureCallback) {
        logger.logMethod()
        takePictureCallbacks.remove(cameraTakePictureCallback)
    }

    override fun setAutoFocusOnTouch(autoFocusOnTouch: Boolean) {
        this.autoFocusOnTouch = autoFocusOnTouch
    }

    override fun autoFocus() {
        tryAutoFocus(null)
    }

    private fun tryAutoFocus(autofocusCallback: AutofocusCallback?) {
        val point = meteringPointFactory?.createPoint(
            (previewView.measuredWidth / 2).toFloat(),
            (previewView.measuredHeight / 2).toFloat()
        )
        point?.let {
            val action = FocusMeteringAction.Builder(point, FocusMeteringAction.FLAG_AF)
                .addPoint(point, FocusMeteringAction.FLAG_AE) // could have many
                // auto calling cancelFocusAndMetering in 5 seconds
                .setAutoCancelDuration(DEFAULT_FOCUS_DURATION, TimeUnit.SECONDS)
                .build()

            val future = camera?.cameraControl?.startFocusAndMetering(action)
            if (cameraExecutor.isShutdown.not()) {
                future?.addListener(Runnable {
                    try {
                        val result = future.get()
                    } catch (e: Exception) {
                        logger.logException(e)
                    } finally {
                        autofocusCallback?.onAutoFocusCompleted()
                    }
                    // process the result
                }, cameraExecutor)
            }
        }
    }

    override fun continuousFocus() {
        camera?.cameraControl?.cancelFocusAndMetering()
    }

    override fun useFlash(useFlash: Boolean) {
        camera?.cameraControl?.enableTorch(useFlash)
    }

    @SuppressLint("SourceLockedOrientationActivity")
    override fun lockToPortrait(lockPicture: Boolean) {
        if (context is Activity) {
            (context as Activity).requestedOrientation = ActivityInfo.SCREEN_ORIENTATION_SENSOR_PORTRAIT
        }
        isOrientationLocked = true
        isOrientationHardLocked = lockPicture
    }

    @SuppressLint("SourceLockedOrientationActivity")
    override fun lockToLandscape(lockPicture: Boolean) {
        if (context is Activity) {
            (context as Activity).requestedOrientation = ActivityInfo.SCREEN_ORIENTATION_SENSOR_LANDSCAPE
        }
        isOrientationLocked = true
        isOrientationHardLocked = lockPicture
    }

    override fun unlockOrientation() {
        isOrientationLocked = false
        isOrientationHardLocked = false
        if (context is Activity) {
            (context as Activity).requestedOrientation = ActivityInfo.SCREEN_ORIENTATION_UNSPECIFIED
        }
        onOrientationChangeListener?.disable()
    }

    override fun usePinchToZoom(usePinchToZoom: Boolean) {
        pinchToZoomEnabled.set(usePinchToZoom)
        setUpPinchToZoom()
    }

    override fun isFlashEnabled(): Boolean {
        return camera?.cameraInfo?.torchState?.value == TorchState.ON
    }

    /**
     * Starts/resumes camera preview stream. Usually used after picture was taken.
     * Note, the preview is started by default only once after the view is attached, so it's required to call it if the preview was stopped and the view was not reattached.
     */
    override fun startPreview() {
        if (::cameraProvider.isInitialized) {
            val preview = previewUseCase
            if (preview == null || !cameraProvider.isBound(preview)) {
                bindCameraUseCases()
            }
        }
    }

    override fun stopPreview() {
        if (::cameraProvider.isInitialized) {
            unbindAll()
        }
    }

    override fun restartPreview() {
        stopPreview()
        startPreview()
    }

    override fun lockMinFocusDistance(lock: Boolean) {
        this.lockMinFocusDistance.set(lock)
    }

    @Throws(IllegalArgumentException::class)
    override fun setOpticalZoomLevel(zoomLevel: Float) {
        if (zoomLevel !in 0.0..1.0) {
            throw CameraConfigurationException("Invalid zoom level value ($zoomLevel)!")
        }

        initialZoomLevel = zoomLevel
        camera?.cameraControl?.setLinearZoom(zoomLevel)
    }

    override fun setOpticalZoomRange(zoomRange: ZoomRange) {
        if (zoomRange.minZoomFactor !in 0.0..1.0
            || zoomRange.maxZoomFactor !in 0.0..1.0
            || zoomRange.minZoomFactor > zoomRange.maxZoomFactor
        ) {
            throw CameraConfigurationException("Invalid zoom range value!")
        }

        this.zoomRange = zoomRange

        camera?.let { camera ->
            val linearZoom: Float = camera.cameraInfo.zoomState.value?.linearZoom ?: 0f
            if (linearZoom < zoomRange.minZoomFactor) {
                setOpticalZoomLevel(zoomRange.minZoomFactor)
            }
            if (linearZoom > zoomRange.maxZoomFactor) {
                setOpticalZoomLevel(zoomRange.maxZoomFactor)
            }
        }
    }

    override fun setPhysicalZoom(zoom: Float) {
        this.physicalZoom = zoom
        camera?.let { camera ->
            val maxZoom = camera.cameraInfo.zoomState.value?.maxZoomRatio ?: 1f
            val minZoom = camera.cameraInfo.zoomState.value?.minZoomRatio ?: 1f
            when {
                zoom <= minZoom -> camera.cameraControl.setZoomRatio(minZoom)
                zoom >= maxZoom -> camera.cameraControl.setZoomRatio(maxZoom)
                else -> {
                    camera.cameraControl.setZoomRatio(zoom)
                }
            }
        }
    }

    override fun setCameraModule(cameraModule: CameraModule) {
        scanbotCameraModule = cameraModule
        cameraXLensFacing = when (cameraModule) {
            CameraModule.BACK -> CameraSelector.LENS_FACING_BACK
            CameraModule.FRONT, CameraModule.FRONT_MIRRORED -> CameraSelector.LENS_FACING_FRONT
        }
        // Reinit supported picture/preview sizes for the new lens facing
        initSupportedPreviewPictureSizes()
    }

    /**
     * Set camera picture size
     *
     * @param pictureSize
     */
    fun setPictureSize(pictureSize: Size) {
        this.customPictureSize = pictureSize
    }

    /**
     * Set camera preview and analyzer frame size
     *
     * @param frameSize
     */
    fun setCameraFrameSize(frameSize: Size) {
        this.customPreviewFrameSize = frameSize
        this.customAnalyzerFrameSize = frameSize
    }

    /**
     * Set camera preview frame size
     *
     * @param frameSize
     */
    fun setPreviewFrameSize(frameSize: Size) {
        this.customPreviewFrameSize = frameSize
    }

    /**
     * Set camera analyzer frame size
     *
     * @param frameSize
     */
    fun setAnalyzerFrameSize(frameSize: Size) {
        this.customAnalyzerFrameSize = frameSize
    }

    /**
     * Set whether camera should force maximum possible size of snapped picture
     * or use maximum size that available with the same aspect ratio as preview
     * by default set to `false`
     * This parameter works only with `ScanbotCameraXView` enabled
     */
    override fun setForceMaxSnappingSize(enabled: Boolean) {
        this.forceMaxSnappingSize = enabled
    }

    private fun setUpPinchToZoom() {
        if (pinchToZoomEnabled.get().not()) {
            scaleGestureDetector = null
            return
        }

        if (scaleGestureDetector == null) {
            val listener = object : ScaleGestureDetector.SimpleOnScaleGestureListener() {
                override fun onScale(detector: ScaleGestureDetector): Boolean {
                    camera?.let { camera ->
                        val zoomRatio: Float = camera.cameraInfo.zoomState.value?.zoomRatio ?: 1f
                        scaleGestureDetector?.let { scaleGestureDetector ->
                            val scaleFactor = scaleGestureDetector.scaleFactor
                            val currentLinearZoom = camera.cameraInfo.zoomState.value?.linearZoom ?: 0f
                            if ((currentLinearZoom <= zoomRange.minZoomFactor && scaleFactor < 1)
                                || (currentLinearZoom >= zoomRange.maxZoomFactor && scaleFactor > 1)
                                || scaleFactor == 1f
                            ) {
                                return true
                            }

                            camera.cameraControl.setZoomRatio(zoomRatio * scaleFactor)

                            val newLinearZoom = camera.cameraInfo.zoomState.value?.linearZoom ?: 0f
                            if (newLinearZoom < zoomRange.minZoomFactor) {
                                camera.cameraControl.setLinearZoom(zoomRange.minZoomFactor)
                                return true
                            }
                            if (newLinearZoom > zoomRange.maxZoomFactor) {
                                camera.cameraControl.setLinearZoom(zoomRange.maxZoomFactor)
                                return true
                            }
                        }
                    }

                    return true
                }
            }

            scaleGestureDetector = ScaleGestureDetector(context, listener)
        }
    }

    private fun initSupportedPreviewPictureSizes() {
        cameraManager?.let { cameraManager ->
            cameraManager.cameraIdList?.find {
                hasSameLensFacing(it)
            }?.let {
                val characteristics = cameraManager.getCameraCharacteristics(it)
                val streamConfigurationMap = characteristics.get(CameraCharacteristics.SCALER_STREAM_CONFIGURATION_MAP)
                supportedPreviewSizes = streamConfigurationMap?.getOutputSizes(SurfaceHolder::class.java)
                    ?: emptyArray()
                supportedPictureSizes = streamConfigurationMap?.getOutputSizes(ImageReader::class.java)
                    ?: emptyArray()
            }
        }
    }

    private fun hasSameLensFacing(cameraId: String): Boolean {
        return cameraManager?.let {
            val cameraCharacteristics = it.getCameraCharacteristics(cameraId)
            val cameraLensFacing = cameraCharacteristics.get(CameraCharacteristics.LENS_FACING)
            cameraLensFacing == this.cameraXLensFacing
        } ?: true
    }

    private fun notifyTakePictureCancelled() {
        takePictureCallbacks.forEach {
            it.onTakePictureCancelled()
        }
    }

    private fun updateTouchRect(success: Boolean) {
        if (touchRect != null) {
            if (touchFocusPaint.color != Color.WHITE) {
                touchFocusPaint.alpha = 0
                invalidate()
                return
            }
            touchFocusPaint.color = ContextCompat.getColor(
                context,
                if (success) android.R.color.holo_green_light else android.R.color.holo_red_light
            )
            invalidate()
            postDelayed({
                if (touchFocusPaint.color != Color.WHITE) {
                    touchFocusPaint.alpha = 0
                    invalidate()
                }
            }, DELAY_HIDE_TOUCH_FOCUS_MS.toLong())
        }
    }

    private fun updateAutoFocusTouchRect(event: MotionEvent) {
        touchFocusPaint.color = ContextCompat.getColor(context, android.R.color.white)
        val x = event.x
        val y = event.y
        touchRect = Rect(
            (x - DEFAULT_FOCUS_AREA_SIZE).toInt(),
            (y - DEFAULT_FOCUS_AREA_SIZE).toInt(),
            (x + DEFAULT_FOCUS_AREA_SIZE).toInt(),
            (y + DEFAULT_FOCUS_AREA_SIZE).toInt()
        )
        invalidate()
    }

    override fun dispatchDraw(canvas: Canvas) {
        super.dispatchDraw(canvas)
        touchRect?.let { canvas.drawRect(it, touchFocusPaint) }
    }

    /**
     * Declare and bind preview, capture and analysis use cases
     */
    @SuppressLint("RestrictedApi", "UnsafeExperimentalUsageError")
    private fun setUpCamera() {
        // Bind the CameraProvider to the LifeCycleOwner
        val cameraProviderFuture = ProcessCameraProvider.getInstance(context)
        cameraProviderFuture.addListener(Runnable {

            // CameraProvider
            cameraProvider = cameraProviderFuture.get()

            bindCameraUseCases()
        }, ContextCompat.getMainExecutor(context))
    }

    @SuppressLint("ClickableViewAccessibility")
    private fun bindCameraUseCases() {
        // Get screen metrics used to setup camera for full screen resolution
        val rotation = previewView.display.rotation

        val cameraSelector = CameraSelector.Builder().requireLensFacing(cameraXLensFacing).build()

        // Must unbind the use-cases before rebinding them
        unbindAll()

        // Preview
        val previewBuilder = Preview.Builder()
            // Set initial target rotation
            .setTargetRotation(rotation)

        setSizeForUseCase(previewBuilder, customPreviewFrameSize, supportedPreviewSizes)

        if (lockMinFocusDistance.get()) {
            enableMinFocusDistance(previewBuilder)
        }

        previewUseCase = previewBuilder.build()

        // ImageCapture
        val imageCaptureBuilder = ImageCapture.Builder()
            .setCaptureMode(ImageCapture.CAPTURE_MODE_MINIMIZE_LATENCY)
            // Set initial target rotation, we will have to call this again if rotation changes
            // during the lifecycle of this use case
            .setTargetRotation(rotation)
            .setIoExecutor(cameraExecutor)

        setSizeForUseCase(imageCaptureBuilder, customPictureSize, supportedPictureSizes)

        imageCaptureUseCase = imageCaptureBuilder.build()

        // ImageAnalysis
        val analyzerBuilder = ImageAnalysis.Builder()
            // Set initial target rotation, we will have to call this again if rotation changes
            // during the lifecycle of this use case
            .setTargetRotation(rotation)
            .setBackpressureStrategy(ImageAnalysis.STRATEGY_KEEP_ONLY_LATEST)
            .setImageQueueDepth(3)

        setSizeForUseCase(analyzerBuilder, customAnalyzerFrameSize, supportedPreviewSizes)

        imageAnalyzer = analyzerBuilder.build()
            // The analyzer can then be assigned to the instance
            .also {
                it.setAnalyzer(analyzerExecutor, compoundImageAnalyzer)
            }

        try {
            // A variable number of use-cases can be passed here -
            // camera provides access to CameraControl & CameraInfo
            lifecycleOwner?.let { lifecycleOwner ->
                camera = cameraProvider.bindToLifecycle(
                    lifecycleOwner, cameraSelector, previewUseCase, imageCaptureUseCase, imageAnalyzer
                )

                // Attach the viewfinder's surface provider to preview use case
                previewUseCase?.setSurfaceProvider(previewView.surfaceProvider)

                if (initialZoomLevel != -1f) {
                    post { camera?.cameraControl?.setLinearZoom(initialZoomLevel) }
                }

                if (physicalZoom != -1f) {
                    post { setPhysicalZoom(physicalZoom) }
                }

                resolution = previewUseCase?.attachedSurfaceResolution?.also { previewSize ->
                    val displayRotation = rotationDegreesFromSurfaceRotation(display.rotation)
                    val isNaturalPortrait = ScaleTypeTransform.isNaturalPortrait(context, display.rotation)

                    // We have to check if the device has a "natural portrait" orientation (usually that means deviceRotation == 0 in the portrait orientation and display width < display height).
                    // And if this flag is `true` and preview width/height doesn't fit to the displayRotation - we have to swap preview width and height.
                    // For example: displayRotation == 0 that corresponds to the portrait orientation and display width < height according to "natural portrait" flag, but current camera preview width > height.
                    // If device is not a natural portrait device - we use current previewSize.
                    // PS: we have do this fix because some devices return the same `previewSize` for portrait and for landscape orientation.
                    val newPreviewSize = if (isNaturalPortrait &&
                        ((displayRotation == 0 || displayRotation == 180) && previewSize.width > previewSize.height)
                        || ((displayRotation == 90 || displayRotation == 270) && previewSize.height > previewSize.width)
                    )
                        Size(previewSize.height, previewSize.width)
                    else previewSize
                    adjustFinderView(newPreviewSize)
                    adjustPolygonView(newPreviewSize)
                }

                meteringPointFactory = PreviewViewMeteringPointFactory(
                    display,
                    camera!!.cameraInfo,
                    resolution,
                    getScaleType(previewMode),
                    width,
                    height
                )

                if (isOrientationLocked
                    && !isOrientationHardLocked
                ) {
                    onOrientationChangeListener?.enable()
                }

                previewView.setOnTouchListener { _, event ->
                    if (pinchToZoomEnabled.get() && event.pointerCount > 1) {
                        continuousFocus()
                        scaleGestureDetector?.onTouchEvent(event)
                    } else {
                        startAutoFocusOnTouch(event)
                    }

                    return@setOnTouchListener true
                }
            }
        } catch (e: Exception) {
            logger.logException(e)
        }
    }

    private fun setSizeForUseCase(builder: ImageOutputConfig.Builder<out Any?>, customSize: Size?, sizes: Array<Size>) {
        if (forceMaxSnappingSize) {
            customSize?.let { builder.setDefaultResolution(it) }
                ?: sizes.maxByOrNull { it.height * it.width }
                    ?.let { builder.setDefaultResolution(it) }
        } else {
            setCalculatedTargetSize(builder, customSize)
        }
    }

    private fun enableMinFocusDistance(useCaseBuilder: Preview.Builder) {
        cameraManager?.let { cameraManager ->
            cameraManager.cameraIdList.find {
                hasSameLensFacing(it)
            }?.let {
                val characteristics = cameraManager.getCameraCharacteristics(it)
                val minFocusDistance: Float = characteristics.get(CameraCharacteristics.LENS_INFO_MINIMUM_FOCUS_DISTANCE) ?: 0f
                val extender = Camera2Interop.Extender(useCaseBuilder)
                extender.setCaptureRequestOption(CaptureRequest.CONTROL_AF_MODE, CameraMetadata.CONTROL_AF_MODE_OFF)
                extender.setCaptureRequestOption(CaptureRequest.LENS_FOCUS_DISTANCE, minFocusDistance)
            }
        }
    }

    @SuppressLint("RestrictedApi")
    private fun setCalculatedTargetSize(builder: ImageOutputConfig.Builder<out Any?>, targetResolution: Size?) {
        if (targetResolution != null) {
            builder.setTargetResolution(targetResolution)
        } else {
            val calculatedPreviewSize = calculatePreviewSize()
            // we have to increase calculated resolution to force Camera to use max default preview and picture size
            builder.setDefaultResolution(
                Size(
                    calculatedPreviewSize.width * PREVIEW_SIZE_MULTIPLIER,
                    calculatedPreviewSize.height * PREVIEW_SIZE_MULTIPLIER
                )
            )
            if (calculatedPreviewSize.width == calculatedPreviewSize.height) {
                // for square Preview we have to increase calculated resolution to force Camera to use max preview size
                builder.setTargetResolution(
                    Size(
                        calculatedPreviewSize.width * PREVIEW_SIZE_MULTIPLIER,
                        calculatedPreviewSize.height * PREVIEW_SIZE_MULTIPLIER
                    )
                )
            } else {
                builder.setTargetAspectRatio(calculateAspectRatio())
            }
        }
    }

    private fun calculateAspectRatio(): Int {
        val calculatePreviewSize = calculatePreviewSize()
        val aspect = calculatePreviewSize.width.toDouble() / calculatePreviewSize.height
        return if (aspect > THRESHOLD_ASPECT_RATIO) AspectRatio.RATIO_16_9 else AspectRatio.RATIO_4_3
    }

    private fun calculatePreviewSize(): Size {
        val measuredWidth = measuredWidth
        val measuredHeight = measuredHeight

        val width = max(measuredWidth, measuredHeight)
        val height = min(measuredWidth, measuredHeight)

        val ratio = width.toDouble() / height

        val listOf = listOf(ASPECT_RATIO_4_3, ASPECT_RATIO_16_9, ASPECT_RATIO_1_1)
        var targetAspect = ASPECT_RATIO_1_1
        listOf.forEach {
            if (abs(it - ratio) < abs(targetAspect - ratio)) {
                targetAspect = it
            }
        }

        return Size(width, (width / targetAspect).toInt())
    }

    private fun unbindAll() {
        cameraProvider.unbindAll()
        previewUseCase = null
        imageCaptureUseCase = null
    }

    private fun startAutoFocusOnTouch(event: MotionEvent) {
        if (event.action != MotionEvent.ACTION_DOWN || !autoFocusOnTouch) {
            return
        }
        updateAutoFocusTouchRect(event)

        val point = meteringPointFactory?.createPoint(event.x, event.y)
        point?.let {
            val action = FocusMeteringAction.Builder(point, FocusMeteringAction.FLAG_AF)
                .addPoint(point, FocusMeteringAction.FLAG_AE) // could have many
                // auto calling cancelFocusAndMetering in 5 seconds
                .setAutoCancelDuration(DEFAULT_FOCUS_DURATION, TimeUnit.SECONDS)
                .build()

            val future = camera?.cameraControl?.startFocusAndMetering(action)
            if (cameraExecutor.isShutdown.not()) {
                future?.addListener(Runnable {
                    previewView.post {
                        try {
                            val result = future.get()
                            updateTouchRect(result.isFocusSuccessful)
                        } catch (e: Exception) {
                            logger.logException(e)
                            updateTouchRect(false)
                        }
                    }
                    // process the result
                }, cameraExecutor)
            }
        }
    }

    private fun getScaleType(previewMode: CameraPreviewMode): PreviewView.ScaleType {
        return when (previewMode) {
            CameraPreviewMode.FIT_IN -> PreviewView.ScaleType.FIT_CENTER
            CameraPreviewMode.FILL_IN -> PreviewView.ScaleType.FILL_CENTER
        }
    }

    private fun adjustPolygonView(previewSize: Size) {
        var polygonView = (this as ViewGroup).children.find { it is BasePolygonView }

        if (polygonView == null) {
            val parent = this.parent as? ViewGroup ?: return
            polygonView = parent.children.find { it is BasePolygonView } ?: return
        }

        correctPolygonViewForPreviewMode(this, polygonView, previewSize)
    }

    private fun correctPolygonViewForPreviewMode(
        container: View,
        polygonView: View, previewSize: Size
    ) {
        // Scale TextureView to fill PreviewView while respecting sensor output size aspect ratio
        val scale = when (previewMode) {
            CameraPreviewMode.FIT_IN -> ScaleTypeTransform.getFitScaleWithBufferAspectRatio(container, previewSize)
            CameraPreviewMode.FILL_IN -> ScaleTypeTransform.getFillScaleWithBufferAspectRatio(container, previewSize)
        }

        val measuredWidth = container.measuredWidth
        val measuredHeight = container.measuredHeight

        val newWidth = measuredWidth * scale.first
        val newHeight = measuredHeight * scale.second

        val newLayoutParams = polygonView.layoutParams as LayoutParams

        newLayoutParams.leftMargin = ((measuredWidth - newWidth) / 2).toInt()
        newLayoutParams.rightMargin = ((measuredWidth - newWidth) / 2).toInt()
        newLayoutParams.topMargin = ((measuredHeight - newHeight) / 2).toInt()
        newLayoutParams.bottomMargin = ((measuredHeight - newHeight) / 2).toInt()
        polygonView.layoutParams = newLayoutParams
    }

    private fun getFinderView(): IFinderView? {
        val parentView = parent as? ViewGroup ?: return null
        var finderOverlay = parentView.findViewById<View>(finderViewId)

        if (finderOverlay == null) {
            val grandParent = parentView.parent as? ViewGroup ?: return null
            finderOverlay = grandParent.findViewById(finderViewId) ?: return null
        }
        return finderOverlay as? IFinderView
    }

    private fun adjustFinderView(previewSize: Size) {
        if (this.parent != null) {
            val finderOverlay = getFinderView()

            finderOverlay?.setCameraParameters(previewSize.width, previewSize.height, previewMode)
        }
    }

    private fun calculateVisibleRectF(previewWidth: Int, previewHeight: Int, previewMode: CameraPreviewMode): RectF {
        if (previewMode == CameraPreviewMode.FIT_IN) {
            return RectF(0f, 0f, 1f, 1f)
        }

        val parentWidth = measuredWidth
        val parentHeight = measuredHeight
        val heightScale = parentHeight / previewHeight.toDouble()
        val widthScale = parentWidth / previewWidth.toDouble()
        val scaledPreviewWidth: Int
        val scaledPreviewHeight: Int

        if (heightScale > widthScale) {
            scaledPreviewWidth = (previewWidth * heightScale).toInt()
            scaledPreviewHeight = parentHeight
        } else {
            scaledPreviewWidth = parentWidth
            scaledPreviewHeight = (previewHeight * widthScale).toInt()
        }
        val previewOffsetWidth = (scaledPreviewWidth - parentWidth) / 2
        val previewOffsetHeight = (scaledPreviewHeight - parentHeight) / 2
        val overlayTop = normalizeF(previewOffsetHeight / scaledPreviewHeight.toFloat())
        val overlayLeft = normalizeF(previewOffsetWidth / scaledPreviewWidth.toFloat())
        val overlayBottom = normalizeF((scaledPreviewHeight - previewOffsetHeight) / scaledPreviewHeight.toFloat())
        val overlayRight = normalizeF((scaledPreviewWidth - previewOffsetWidth) / scaledPreviewWidth.toFloat())

        return RectF(overlayLeft, overlayTop, overlayRight, overlayBottom)
    }

    private fun calculateFinderRectF(previewWidth: Int, previewHeight: Int, previewMode: CameraPreviewMode): RectF? {
        if (this.parent != null) {
            val finderOverlay = getFinderView() as? FinderOverlayView
            if (finderOverlay == null || finderOverlay.visibility != View.VISIBLE) {
                return null
            }
            val parent = finderOverlay.parent as ViewGroup
            val parentWidth = parent.measuredWidth
            val parentHeight = parent.measuredHeight

            var overlayRect = Rect()
            finderOverlay.getDrawingRect(overlayRect)
            parent.offsetDescendantRectToMyCoords(finderOverlay, overlayRect)
            overlayRect = Rect(
                overlayRect.left + finderOverlay.overlayLeftPadding,
                overlayRect.top + finderOverlay.overlayTopPadding,
                overlayRect.right - finderOverlay.overlayRightPadding,
                overlayRect.bottom - finderOverlay.overlayBottomPadding
            )
            val heightScale = parentHeight / previewHeight.toDouble()
            val widthScale = parentWidth / previewWidth.toDouble()
            val scaledPreviewWidth: Int
            val scaledPreviewHeight: Int
            if (previewMode == CameraPreviewMode.FILL_IN) {
                if (heightScale > widthScale) {
                    scaledPreviewWidth = (previewWidth * heightScale).toInt()
                    scaledPreviewHeight = parentHeight
                } else {
                    scaledPreviewWidth = parentWidth
                    scaledPreviewHeight = (previewHeight * widthScale).toInt()
                }
            } else {
                if (heightScale > widthScale) {
                    scaledPreviewWidth = parentWidth
                    scaledPreviewHeight = (previewHeight * widthScale).toInt()
                } else {
                    scaledPreviewWidth = (previewWidth * heightScale).toInt()
                    scaledPreviewHeight = parentHeight
                }
            }
            val previewOffsetWidth = (scaledPreviewWidth - parentWidth) / 2
            val previewOffsetHeight = (scaledPreviewHeight - parentHeight) / 2
            val overlayTop = normalizeF((previewOffsetHeight + overlayRect.top) / scaledPreviewHeight.toFloat())
            val overlayLeft = normalizeF((previewOffsetWidth + overlayRect.left) / scaledPreviewWidth.toFloat())
            val overlayBottom = normalizeF((previewOffsetHeight + overlayRect.bottom) / scaledPreviewHeight.toFloat())
            val overlayRight = normalizeF((previewOffsetWidth + overlayRect.right) / scaledPreviewWidth.toFloat())
            return RectF(overlayLeft, overlayTop, overlayRight, overlayBottom)
        }
        return null
    }

    private fun normalizeF(rectCoordinates: Float): Float {
        return when {
            rectCoordinates < 0 -> 0f
            rectCoordinates > 1 -> 1f
            else -> rectCoordinates
        }
    }

    /**
     * Compound image analysis class.
     *
     * All we need to do is override the function `analyze` with our desired operations.
     */
    private inner class CompoundAnalyzer : ImageAnalysis.Analyzer {
        private val frameHandlers = LinkedHashSet<FrameHandler>()

        private val framesInProcessing = AtomicInteger(0)
        private val isUnderBackpressure: Boolean
            get() = framesInProcessing.get() >= MAX_FRAMES_IN_PROCESSING

        /**
         * Check whether there is same class of [FrameHandler] attached to the view
         * return object of the same class if exists otherwise [null]
         */
        fun <T : FrameHandler> getAttachedFrameHandler(clazz: Class<T>): T? {
            return frameHandlers.firstOrNull { it::class.java == clazz } as? T
        }

        /**
         * Used to add frame handler that will be called on each next frame
         */
        fun addFrameHandler(frameHandler: FrameHandler) {
            synchronized(frameHandlers) {
                frameHandlers.add(frameHandler)
            }
        }

        /**
         * Unregisters [FrameHandler].
         *
         * @param frameHandler handler to unregister. If it was not registered before, nothing happens.
         */
        fun removeFrameHandler(frameHandler: FrameHandler) {
            synchronized(frameHandlers) {
                frameHandlers.remove(frameHandler)
            }
        }

        /**
         * Analyzes an image to produce a result.
         *
         * The caller is responsible for ensuring this analysis method can be executed quickly
         * enough to prevent stalls in the image acquisition pipeline. Otherwise, newly available
         * images will not be acquired and analyzed.
         *
         * The image passed to this method becomes invalid after this method returns. The caller
         * should not store external references to this image, as these references will become
         * invalid.
         *
         * @param image image being analyzed VERY IMPORTANT: Analyzer method implementation must
         * call image.close() on received images when finished using them. Otherwise, new images
         * may not be received or the camera may stall, depending on back pressure setting.
         *
         */
        @SuppressLint("RestrictedApi")
        override fun analyze(image: ImageProxy) {
            // If there are no listeners attached, we don't need to perform analysis
            if (isUnderBackpressure || frameHandlers.isEmpty()) {
                image.close()
                return
            }
            framesInProcessing.incrementAndGet()

            val imageOrientation: Int = image.imageInfo.rotationDegrees
            val imageWidth = image.width
            val imageHeight = image.height

            // Extract image data from callback object
            val data = convertYUV420toNV21(image)
            image.close()

            val finalFrame = if (scanbotCameraModule == CameraModule.FRONT_MIRRORED) {
                mirrorBufferFrame(data, imageWidth, imageHeight, imageOrientation)
            } else data

            if (frameAnalyzerExecutor.isShutdown) {
                return
            }
            frameAnalyzerExecutor.execute {
                var finderRect: Rect? = null
                var visibleRectF: RectF? = RectF(0f, 0f, 1f, 1f)

                val previewWidth: Int
                val previewHeight: Int

                if (imageOrientation == 0 || imageOrientation == 180) {
                    previewWidth = if (imageWidth < imageHeight) imageHeight else imageWidth
                    previewHeight = if (imageWidth < imageHeight) imageWidth else imageHeight
                } else {
                    previewWidth = if (imageWidth > imageHeight) imageHeight else imageWidth
                    previewHeight = if (imageWidth > imageHeight) imageWidth else imageHeight
                }
                val overlayRectF: RectF? = calculateFinderRectF(previewWidth, previewHeight, previewMode)
                visibleRectF = calculateVisibleRectF(previewWidth, previewHeight, previewMode)
                if (overlayRectF != null) {
                    finderRect = Rect(
                        (previewWidth * overlayRectF.left).toInt(),
                        (previewHeight * overlayRectF.top).toInt(),
                        (previewWidth * overlayRectF.right).toInt(),
                        (previewHeight * overlayRectF.bottom).toInt()
                    )
                    if (finderRect.left < 0 || finderRect.left > previewWidth || finderRect.top < 0 || finderRect.top > previewHeight || finderRect.right < 0 || finderRect.right > previewWidth || finderRect.bottom < 0 || finderRect.bottom > previewHeight || finderRect.width() <= 0 || finderRect.height() <= 0) {
                        logger.e("Camera", "Finder overlay layout does not fit to the Camera preview frame.")
                        framesInProcessing.decrementAndGet()
                        return@execute
                    }

                    currentFinderRectF = listOf(
                        PointF(overlayRectF.left, overlayRectF.top),
                        PointF(overlayRectF.right, overlayRectF.top),
                        PointF(overlayRectF.right, overlayRectF.bottom),
                        PointF(overlayRectF.left, overlayRectF.bottom)
                    )
                } else {
                    currentFinderRectF = emptyList()
                }

                var handled = false
                // Call all listeners with new value
                frameHandlers.forEach {
                    if (handled) {
                        return@forEach
                    }
                    val result = it.handleFrame(
                        FrameHandler.Frame(
                            finalFrame,
                            imageWidth,
                            imageHeight,
                            imageOrientation,
                            finderRect,
                            visibleRectF,
                            finderInnerThresholdPx,
                            finderOuterThresholdPx
                        )
                    )
                    handled = result
                }
                framesInProcessing.decrementAndGet()
            }
        }

        private fun mirrorBufferFrame(frame: ByteArray, frameWidth: Int, frameHeight: Int, imageOrientation: Int): ByteArray {
            val frameBitmap = PreviewImageUtil.convertNV21ToBitmap(frame, frameWidth, frameHeight, 0)
            val matrix = Matrix()
            if (imageOrientation == 0 || imageOrientation == 180) {
                matrix.preScale(-1f, 1f)
            } else {
                matrix.preScale(1f, -1f)
            }
            val newBitmap: Bitmap = Bitmap.createBitmap(
                frameBitmap, 0, 0, frameWidth,
                frameHeight, matrix, true
            )

            val finalFrame = CameraImageUtils.bitmapToNv21(newBitmap, frameWidth, frameHeight)
            frameBitmap.recycle()
            newBitmap.recycle()
            return finalFrame
        }
    }

    private inner class OnOrientationChange(context: Context?) :
        OrientationEventListener(context, SensorManager.SENSOR_DELAY_FASTEST) {
        private var currentOrientation = ORIENTATION_UNKNOWN
        private var isEnabled = false

        init {
            disable()
        }

        override fun onOrientationChanged(orientation: Int) {
            var orientationNew = orientation
            if (camera == null || !canDetectOrientation() || orientationNew == ORIENTATION_UNKNOWN) {
                return
            }
            orientationNew = getClosestRotationDegree(orientationNew)
            if (orientationNew != currentOrientation) {
                val rotation = when (orientationNew) {
                    0 -> Surface.ROTATION_0
                    90 -> Surface.ROTATION_270
                    180 -> Surface.ROTATION_180
                    270 -> Surface.ROTATION_90
                    else -> Surface.ROTATION_0
                }
                logger.d(TAG, "Rotation changed: $rotation")
                imageCaptureUseCase?.targetRotation = rotation
                imageAnalyzer?.targetRotation = rotation
                currentOrientation = orientationNew
            }
        }

        private fun getClosestRotationDegree(rotation: Int): Int {
            for (value in ROTATION_DEGREES) {
                val diff = abs(rotation - value)
                if (diff < 45) {
                    return value
                }
            }
            return 0
        }

        override fun enable() {
            isEnabled = true
            super.enable()
        }

        override fun disable() {
            isEnabled = false
            super.disable()
        }

        fun isEnabled(): Boolean {
            return isEnabled
        }

        fun resetOrientation() {
            currentOrientation = ORIENTATION_UNKNOWN
        }
    }

    /**
     * Wrapper needs to call each CameraStateCallback from lifecycle directly
     */
    private inner class CameraStateCallbackWrapper(val stateCallback: CameraStateCallback) : LifecycleObserver {

        @OnLifecycleEvent(Lifecycle.Event.ON_START)
        fun onStart(lifecycleOwner: LifecycleOwner) {
            stateCallback.onResume()
        }

        @OnLifecycleEvent(Lifecycle.Event.ON_STOP)
        fun onStop(lifecycleOwner: LifecycleOwner) {
            stateCallback.onPause()
        }

        @OnLifecycleEvent(Lifecycle.Event.ON_DESTROY)
        fun onDestroy(lifecycleOwner: LifecycleOwner) {
            lifecycleOwner.lifecycle.removeObserver(this)
        }
    }

    /**
     * Listens for auto-focus callbacks
     */
    private interface AutofocusCallback {
        fun onAutoFocusCompleted()
    }

    override fun setShutterSound(value: Boolean) {
        // not supported
    }

    companion object {
        private const val TAG = "ScanbotCameraXView"

        private const val MAX_FRAMES_IN_PROCESSING = 1

        private const val DELAY_HIDE_TOUCH_FOCUS_MS = 1000
        private const val DEFAULT_FOCUS_AREA_SIZE = 75
        private const val DEFAULT_FOCUS_DURATION = 5L

        private const val THRESHOLD_ASPECT_RATIO = 5 / 3.0
        private const val ASPECT_RATIO_16_9 = 16 / 9.0
        private const val ASPECT_RATIO_4_3 = 4 / 3.0
        private const val ASPECT_RATIO_1_1 = 1.0
        private const val PREVIEW_SIZE_MULTIPLIER = 100

        private val ROTATION_DEGREES = intArrayOf(0, 90, 180, 270)
    }

    override fun setUiZoomLevel(zoomLevel: Float) {
        setPhysicalZoom(zoomLevel)
    }
}
